#!/usr/bin/perl -w
#
# quizz CGI script
#
# Thomas Linden <tom@co.daemon.de>
#
# Copyright (c) 2003 Thomas Linden.
# ALL RIGHTS RESERVED.
#
# Licensed under the terms of the artistic license.
#
# The module "Template" is Copyright 1999-2000 Jens Heunemann.
# Used with permission.
#


use CGI qw/:standard/;
use Data::Dumper;
use CGI::Carp qw(fatalsToBrowser);
use CGI::Cookie;
use Template;
use Config::General;
use DB_File;
use Digest::MD5 qw(md5_base64);


# initialize vars, change as appropriate
my $cgi            = "index.cgi";
my $path_conf      = "config";
my $path_data      = "data";
my $path_templates = "templates";


# read/parse config file using Config::General
my %config = ParseConfig(
			 -ConfigFile       => "$path_conf/quizz.cfg",
			 -InterPolateVars  => 1,
			 -UseApacheInclude => 1,
			 -IncludeRelative  => 1
			   );

my %choice = (
	      M => "multiple",
	      O => "multiple",
	      S => "single",
	      R => "regex"
	      );

my $VERSION  = "1.1.3";

my @loop = @{$config{loop}};

# cgi object
my $query    = new CGI;

# result database (dbm)
my %db;

# initialize if not already done so far
&prepare;

# initialize shared templates
my $head     = new Template("$path_templates/header.tpl");
my $foot     = new Template("$path_templates/footer.tpl");

# initialize cgi parameters
my %input = map { $_ => $query->param($_) } $query->param;

# check it up where to go
my $sub2go   = $input{state} || "default";




eval {
  # now do the real work
  &$sub2go;
};
if ($@) {
  print "Content-Type: text/html\n\n\n";
  print $@;
}
if ($config{debug}) {
  print "<pre>\n" . Dumper(\%input) . "\n</pre>";
}

# and the foot including the last changed date
my @stat = lstat($cgi);
$foot->code(
	    date => scalar localtime( $stat[9] ),
	    version => $VERSION
	   );
$foot->print;

exit;





###### all subs from now on, called by ::main via closure call (&$sub2go) #####
sub dump {
  my $var = shift;
  print "<pre>\n" . Dumper($var) . "\n</pre>";
}

sub default {
  print "Content-Type: text/html\n\n\n";
  $head->print;
  my $T = new Template(-file => "$path_templates/default.tpl");

  # compute best-of list
  my @_bestof;
  foreach my $user (keys %db) {
    my %entry = &get_entry($user);
    push @_bestof, \%entry;
  }
  my $num = 1;
  foreach my $entry (sort { $b->{points} <=>  $a->{points} } @_bestof) {
    if ($num <= $config{config}->{max_best}) {
      $entry->{pos} = $num;
      $entry->{color} = $loop[0];
      @loop = reverse @loop;
      push @bestof, $entry;
    }
    else {
      last;
    }
    $num++;
  }

  $T->code(
	   data => {
		    quiz_name        => $config{quiz_name},
		    quiz_description => $config{quiz_description},
		    cgi              => $cgi
		   },
	   bestof => \@bestof
	  );

  print $T->process;
}

sub quest {
  if ($input{user} eq "") {
    &default;
    return;
  }

  if (! exists $input{question}) {
    # starting point
    $input{question} = 1;
    my $session;

    if ($config{config}->{cookie} ne "none") {
      # check cookie
      my %cookie = fetch CGI::Cookie;
      my ($name, $md5) = &get_md5;
      if (exists $cookie{$name}) {
	$session = $cookie{$name}->value();
	if ($md5 == $session) {
	  &default; # not allowed!
	  print "<I>no retry as user <b>$input{user}</b> possible.</I>\n";
	  return;
	}
      }
      else {
	# create new cookie
	my $c = new CGI::Cookie(
				-name    => $name,
				-value   => $md5,
				-expires => $config{config}->{cookie_expire} || '+1M',   # expires after 1 week
			       );
	print header(-cookie=>[$c]);
	$head->print;
      }
    }
    else {
      # no cookies, but dont forget to send the header
      print "Content-Type: text/html\n\n\n";
      $head->print;
    }
  }
  else {
    # previous run was a question
    print "Content-Type: text/html\n\n\n";
    $head->print;

    my %q = %{ $config{question}->{ $input{question} } };
    my $factor = $q{points} / $q{time};
    my %entry = &get_entry($input{user});
    my ($points, $time, $result);

    if ($input{question} == 1) {
      $entry{points} = 0;
      $entry{time}   = 0;
    }

    # calculate time and points based on time
    # points will be zeroed if answer is wrong!
    if (exists $input{time}) {
      # time counter
      $points = ( int($input{time} * $factor) ) + $entry{points};
      $time   = $input{time} + $entry{time};
    }
    else {
      # point counter
      $points = $input{points} + $entry{points};
      $time   = ( int($input{points} / $factor) ) + $entry{time};
    }
    $result = "correct";

    # compute the answer
    if ($q{type} eq "M" || $q{type} eq "O") {
      # multiple/optional-multiple choice
      my %choices = map { $_ => 1 } $query->param("choice");
      my %correct = map { $_ => 1 } split /,/, $q{result};
      my $matches;
      foreach my $choice (keys %correct) {
	if (exists $choices{$choice}) {
	  $matches++;
	}
      }
      if ($q{type} eq "M") {
	if ($matches != scalar keys %correct) {
	  # shit, wrong answer!
	  $points = $entry{points};
	  $result = "wrong";
	}
      }
      else {
	# $q{type} eq "O"
	if ($matches < $q{required}) {
	  # shit, wrong answer!
	  $points = $entry{points};
	  $result = "wrong";
	}
      }
    }
    elsif ($q{type} eq "S") {
      # single choice
      if ($input{choice} ne $q{result}) {
	# shit, wrong answer!
	$points = $entry{points};
	$result = "wrong";
      }
    }
    elsif ($q{type} eq "R") {
      # regex choice
      if ($input{choice} !~ /$q{result}/) {
	# shit, wrong answer!
	$points = $entry{points};
	$result = "wrong";
      }
    }

    print "<p><b>Your last answer was: $result</b></p>\n";

    &add_entry(time => $time, points => $points, user => $input{user}, date => scalar localtime(time));
    $input{question}++; # use next one
  }

  if (exists $config{question}->{ $input{question} }) {
    my $T = new Template(-file => "$path_templates/question.tpl");
    my %question = %{ $config{question}->{ $input{question} } };

    if ($config{config}->{counter_time}) {
      $T->code(
	       timer => {
			 max      => $question{time},
			 question => $question{question}
			}
	      );
    }

    if ($config{config}->{counter_points}) {
      $T->code(
	       pointer => {
			   max      => $question{points},
			   interval => $question{points} / $question{time},
			   question => $question{question}
			  }
	      );
    }

    # prepare choices
    my @answers;
    if ($question{type} eq "R") {
      my %c = ( choice => "Answer:");
      push @answers, \%c;
    }
    else {
      foreach my $choice (sort {$a cmp $b } keys %{$question{choice}}) {
	my %c = (
		 choice => $choice,
		 answer => $question{choice}->{$choice}->{answer}
		);
	push @answers, \%c;
      }
    }

    $T->code(
	     $choice{ $question{type} } => \@answers,
	     form => {
		      question => $input{question},
		      user     => $input{user},
		      session  => $session,
		      cgi      => $cgi
		     }
	    );

    $T->code( data => { question => $input{question} } );

    print $T->process;
  }
  else {
    # no more questions, present the result
    my %e = &get_entry($input{user});
    my $T = new Template(-file => "$path_templates/result.tpl");
    $T->code( data => {%e, cgi => $cgi} );
    print $T->process;
  }
}








### helper subs ###
sub prepare {
  if (! -d "$path_data/$config{key}") {
    system "mkdir $path_data/$config{key}" and die "Could not prepare application!";
  }
  tie %db, "DB_File", "$path_data/$config{key}/results.db"
    or die "Could not open database $path_data/$config{key}/results.db: $!";
}

sub get_entry {
  my $user = shift;
  my ($points, $time, $date) = split /\//, $db{$user};
  my %entry = (
	       user   => $user,
	       date   => $date,
	       points => $points,
	       time   => $time
	      );
  return %entry;
}

sub add_entry {
  my %e = @_;
  $db{ $e{user} } = "$e{points}/$e{time}/$e{date}";
}


sub get_md5 {
  my ($name, $src);
  if ($config{config}->{cookie} eq "user") {
    $src  = $input{user};
    $name = "$input{user}-quizzsession";
  }
  elsif ($config{config}->{cookie} eq "ip") {
    $src  = $ENV{REMOTE_ADDR};
    $name = "$ENV{REMOTE_ADDR}-quizzsession";
  }
  elsif ($config{config}->{cookie} eq "both") {
    $input{user} . $ENV{REMOTE_ADDR};
    $name = "$input{user}-$ENV{REMOTE_ADDR}-quizzsession";
  }
  my $md5 = md5_base64($src);
  $q = $config{quiz_name};
  $q =~ s/\s//g;
  $name .= "-$q";
  return ($name, $md5);
}
